/*********************************************************************/
/*      Licensed Materials - Property of IBM                         */
/*      "Restricted Materials of IBM"                                */
/*      5748-T15                                                     */
/*      (C) Copyright IBM Corp. 2006                                 */
/*      All Rights Reserved                                          */
/*      US Government Users Restricted Rights - Use, duplication     */
/*      or dislosure restricted by GSA ADP Schedule Contract         */
/*      with IBM Corp.                                               */
/*********************************************************************/
/* This file is provided on an "as is" basis.  There are no          */
/* warranties, express or implied, including the implied warranties  */
/* of merchantability and fitness for a particular purpose.          */
/* IBM has no obligation to provide service, defect correction, or   */
/* any maintenance for the file.  IBM has no obligation to supply    */
/* any updates or enhancements for the file to you even if such are  */
/* or later become available.                                        */
/*********************************************************************/
/* Change History                                                    */
/*                                                                   */
/*  d20227   20100525  Add messages when database not initialized.   */
/*                     (msgs copied from coro version)               */
/*                                                                   */
/*                                                                   */
/*********************************************************************/
/*                                                                   */
/* GENERAL INFORMATION:                                              */
/*                                                                   */
/* Segment Name:  QMS2                                               */
/*                                                                   */
/*   Load Module Name:  QMS2                                         */
/*                                                                   */
/* Abstract:     commit/rollback test driver.                        */
/*                                                                   */
/* Description:  commit rollback test driver is activated via        */
/*               ZTEST CO64 command.                                 */
/*                                                                   */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <tpf/tpfapi.h>
#include <tpf/tpfio.h>
#include <tpf/tpfglbl.h>
#include <c_coro.h>
#include <tpf/i_tmcr.h>
#include <c_sd0rv.h>
#include <tpf/c_decb.h>

#define printmsg if ( sd0rv->sd0flg3 && SD03PRT ) printf

typedef enum {COMMIT,ROLLBACK} ACTION;

TESTBLK *find_root(int);
int  get_value(void);
void create_fa (void);
void update_fa(void);
int  run_tx (int, int);
void release_fa (void);
void verify_fa (int, char* rid);

struct SD0RV *sd0rv;

/*-------------------------------------------------------------------*/
/* Function    : almost main                                         */
/*                                                                   */
/* Description : For START command :                                 */
/*                 - lock root record on DE                          */
/*                 - add file pool address to the chain if maximum   */
/*                   is not reached yet. (fcnt may exceed the        */
/*                   maximum if ALTER command is issued in between)  */
/*                 - update the file chain, fcnt and chksum          */
/*                 - verify the data content, file chain and chksum. */
/*                 - unhold the root.                                */
/*               For VERIFY command, verify all 200 fixed files :    */
/*                 - lock root in core,                              */
/*                 - verrify the file count and chksum (NO VALUE chk)*/
/*                 - release file chain and init the root block.     */
/*                 - unhold the root.                                */
/*-------------------------------------------------------------------*/
extern "C" void QMS2(void)
{
   TESTBLK        *root;
   int             lvl = 0, ord, value;

   printmsg("entering QMS2 \n");

   cinfc(CINFC_WRITE,CINFC_CMMCRT);

   sd0rv = (struct SD0RV*) (long)* ((int *) glob(_issdrv));

   ord = *( (int *) &ecbptr()->ebw000 );

   if ( ord )   /* START command */
   {
       root = find_root(ord);
       value = root->value;                /* hold   */
       file_record(DE, 0, "CR", 0, NOHOLD);

       run_tx(lvl,value);
       sd0rv->sd0aecb --;

       unfrc(DE);                          /* unhold */
   }
   else  /* VERIFY command, check all 200 fixed file addresses */
   {
       for ( ord = 1; ord <= 200; ord ++ )
       {
           root = find_root(ord);          /* hold */
           if ( root->bwd )                /* fix file in use */
           {
              file_record(DE, 0, "CR", 0, NOHOLD);
              /* find_root set up RID in DF, SD0RV is not set */
              verify_fa(0, (char *)&(ecbptr()->ce1faf));
              release_fa();                /* rlcha  */
           }
           else
           {
              file_record(DE, 0, "CR", 0, NOHOLD);
           }

           unfrc(DE);                      /* unhold */
       }
       wtopc("CORO64 driver database verification is completed",
          WTOPC_EBROUT + WTOPC_PRC + WTOPC_RO, WTOPC_NO_CHAIN, NULL);
   }

   printmsg("exiting QMS2 \n");

   exit(1);
}

/*-------------------------------------------------------------------*/
/* Function    : run_tx                                              */
/*                                                                   */
/* Description : 1. start a transaction                              */
/*               2. create file chain                                */
/*               3. update file chain                                */
/*               4. start a nested transaction if required           */
/*               5. verify file chain content of current transaction */
/*               6. stop the current level transaction               */
/*-------------------------------------------------------------------*/
int  run_tx(int lvl, int value)
{
   int            i;
   int            act;

   lvl ++;

   if ( lvl <= sd0rv->sd0nest )
   {
     for ( act = (int) COMMIT; act <= (int) ROLLBACK; act ++ )
     {
         printmsg("TX_BEGIN %d \n", lvl);
         tx_begin();

         value ++;

         create_fa();
         update_fa();

         value = run_tx(lvl, value);

         if ( act == (int) COMMIT )
         {
            printmsg("TX_COMMIT %d \n",lvl);
            tx_commit();
            sd0rv->sd0cmt ++;
         }
         else
         {
            printmsg("TX_ROLLBACK %d \n",lvl);
            sd0rv->sd0rlb ++;
            for ( i = lvl; i <= sd0rv->sd0nest; i ++)
            {
                value --;
            }
            tx_rollback();
         }

         verify_fa(value, (char *)(&sd0rv->sd0riat));
     }
  }

  return value;
}

/*-----------------------------------------------------------------*/
/* Function    : find_root                                         */
/*                                                                 */
/* Description : get a fixed file address as the root of file chain*/
/*               QFN1 pass us the ordinal number in EBW000         */
/*-----------------------------------------------------------------*/
TESTBLK *find_root(int ord)
{
   TESTBLK          *root;
   int               rc;
   unsigned int     retord;

   if ( (rc = face_facs(ord, "#QL0QL", 0, DE, &retord)) != 0 )
   {
      printf("face rc = %d \n",rc);
      serrc_op_ext(SERRC_EXIT,0xCCCC,"FACS error in finding root "
                   "record",'I',NULL);
   }

   if ( (root = (TESTBLK *)find_record(DE, 0, "CR", 0, HOLD)) == 0)
   {
      printf("CORO DRIVER record may not be initialized\n");
      printf("Please issue ZTEST CORO RESET and then restart the driver\n");

      serrc_op_ext(SERRC_EXIT,0xCCCC,"find error in finding root "
                   "record",'I',NULL);
   }

   if ( *( (int *) &ecbptr()->ebw000) )           /* for start cmd */
   {
       root->bwd = ord;
       memcpy((char *)&root->bwd,sd0rv->sd0riat,2);
   }
   else                                           /* for verify cmd */
   {
       memcpy((char *)&(ecbptr()->ce1faf),(char *)&root->bwd,2);
   }

   memset(root->tid,0x00,sizeof(GTRID));          /* clear out TID */

/* file_record(DE, 0, "CR", 0, NOHOLD); */

   return root;
}

/*-----------------------------------------------------------------*/
/* Function    : create_fa                                         */
/*                                                                 */
/* Description : . get file pool chain to top of queue,            */
/*               . update fcnt in the root,                        */
/*               . chksum is not chanaged now.                     */
/*                                                                 */
/*               . Create a DECB, locate before using it, swapblock*/
/*                 back and forth, delete by name.                 */
/*-----------------------------------------------------------------*/
void create_fa(void)
{
   TESTBLK               *root, *blk;
   TPF_FA8                pool8;
   unsigned int           pool;
   int                    i;
   TPF_DECB              *decb = NULL;
   char                   decb_name[16] = "CORO create_fa";

   if ( (root = (TESTBLK *)find_record(DE, 0, "CR", 0, NOHOLD)) == 0)
   {
      serrc_op_ext(SERRC_EXIT,0xCCCC,"error finding prime record in"
                   "create function",'I',NULL);
   }

   if ( root->fcnt < sd0rv->sd0nest * sd0rv->sd0fcnt )
   {
       if ( (decb = tpf_decb_create(decb_name, NULL)) == NULL)
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error creating DECB",
                        'I',NULL);
       }

      for ( i = 0; i < sd0rv->sd0fcnt; i ++ )
      {
         pool8 = getfc(decb,GETFC_TYPE0,sd0rv->sd0riat,
                      (int)(GETFC_BLOCK+GETFC_FILL),GETFC_SERRC,0x00);
         memcpy(&pool, ((int *)(&pool8) + 1), sizeof(int));
         printmsg("getfc = %x \n", pool);

         tpf_decb_swapblk(decb, DF);           /* test swap block */
         blk = (TESTBLK *) ECB_CBRW_CA(DF);
         tpf_decb_swapblk(decb, DF);           /* swap it back    */

         memcpy(blk->rid,sd0rv->sd0riat,2);
         blk->rcc = 0;
         blk->bwd = 0;
         blk->value = 0;

         blk->fwd = root->fwd;

         root->fwd = pool;
         root->fcnt ++;

         if ( (decb = tpf_decb_locate(decb_name,NULL)) == NULL )
         {
             serrc_op_ext(SERRC_EXIT,0xCCCC,"error locating DECB",
                          'I',NULL);
         }
         file_record_ext(decb, 0, sd0rv->sd0riat, 0, NOHOLD,
                         FILE_DEFEXT);
      }

      tpf_decb_release(decb_name);
   }

   memcpy(root->tid,(CE2TMCR)->tid,sizeof(GTRID));
   file_record(DE, 0, "CR", 0, NOHOLD);
   return;
}

/*-----------------------------------------------------------------*/
/* Function    : update_fa                                         */
/*                                                                 */
/* Description : update the entire FA_CHAIN with value.            */
/*               update chksum in the root.                        */
/*               cretae a DECB for find and file                   */
/*-----------------------------------------------------------------*/
void update_fa(void)
{
   TESTBLK               *blk, *root;
   unsigned int           next;
   TPF_FA8                fa8 = NULL;
   TPF_DECB              *decb = NULL;

   if ( (root = (TESTBLK *)find_record(DE, 0, "CR", 0, NOHOLD)) == 0)
   {
      serrc_op_ext(SERRC_EXIT,0xCCCC,"error finding prime record in"
                   "update function",'I',NULL);
   }

   root->value ++;
   root->chksum = 0;
   memcpy(root->tid,(CE2TMCR)->tid,sizeof(GTRID));

   next = root->fwd;

   while ( next )
   {
       printmsg("chain = %x \n", next);

       if ( (decb = tpf_decb_create(NULL, NULL)) == NULL)
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error creating DECB",
                        'I',NULL);
       }

       memcpy(((int *)&fa8 + 1), &next, sizeof(int));
       if ((blk = (TESTBLK *)find_record_ext(decb,
                                            &fa8,
                                            sd0rv->sd0riat,
                                            0,
                                            NOHOLD_WAIT,
                                            FIND_DEFEXT)) == NULL)
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error finding chain record"
                        " in update function",'I',NULL);
       }
       blk->value = root->value;
       memcpy(blk->tid,(CE2TMCR)->tid,sizeof(GTRID));
       root->chksum += blk->value;

       next = blk->fwd;

       file_record_ext(decb, 0, sd0rv->sd0riat, 0, NOHOLD,FILE_DEFEXT);
       tpf_decb_release(decb);
    }

    file_record(DE, 0, "CR", 0, NOHOLD);

    return;
}

/*-----------------------------------------------------------------*/
/* Function    : verify_fa                                         */
/*                                                                 */
/* Description : verify the data content.                          */
/*               1. value we kept in program is the same as file,  */
/*                  (only valid during run time, VERFIY command    */
/*                  will bypass this check by passing value 0)     */
/*               2. number of file chain = fcnt field in the root. */
/*               3. sum of value in file chain = chksum field in   */
/*                  the root.                                      */
/*                                                                 */
/*               Create, locate and release DECB for each find/file*/
/*                                                                 */
/*-----------------------------------------------------------------*/
void verify_fa(int value, char *rid)
{
   TESTBLK               *blk, *root;
   unsigned int           next;
   int                    fcnt, chksum = 0;
   DECBC_RC               rc;
   TPF_FA8                fa8 = NULL;
   TPF_DECB              *decb = NULL;
   char                   decb_name[16] = "coro verify_fa";

   if ( (root = (TESTBLK *)find_record(DE, 0, "CR", 0, NOHOLD)) == 0)
   {
      serrc_op_ext(SERRC_EXIT,0xCCCC,"error finding prime record in"
                   " verify function",'I',NULL);
   }

   next = root->fwd;
   fcnt = root->fcnt;

   while ( next )
   {
       if ( (decb = tpf_decb_create(decb_name, NULL)) == NULL)
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error creating DECB",
                        'I',NULL);
       }

       memcpy(((unsigned int *)&fa8 + 1), &next, sizeof(int));
       if ((blk = (TESTBLK *)find_record_ext(decb,
                                             &fa8,
                                             rid,
                                             0,
                                             NOHOLD_WAIT,
                                             FIND_DEFEXT))
                                             == 0 ) /* RID on DF */
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error finding chain record"
                        " in verify function",'I',NULL);
       }
       if ( value && blk->value != value )
       {
          serrc_op_ext(SERRC_EXIT,0xCCCC,"CORO driver value verfication"
                       " failed",'I',NULL);
       }

       fcnt --;
       chksum += blk->value;

       next = blk->fwd;
       if ( (decb = tpf_decb_locate(decb_name, NULL)) == NULL )
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error locating DECB",
                        'I',NULL);
       }

       decb = tpf_decb_locate(decb_name, &rc);
       if ( rc != DECBC_OK )
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error locating DECB",
                        'I',NULL);
       }
       file_record_ext(decb, 0, 0, 0, NOHOLD,FILE_DEFEXT);
                                           /* RID should be on DF */

       if ( tpf_decb_validate(decb) != DECBC_OK )
       {
           serrc_op_ext(SERRC_EXIT,0xCCCC,"error validating DECB",
                        'I',NULL);
       }

       tpf_decb_release(decb);
   }

   if ( fcnt )
   {
       serrc_op_ext(SERRC_EXIT,0xCCCC,"CORO driver value file_count "
                    " verification failed",'I',NULL);
   }

   if ( chksum != root->chksum )
   {
       serrc_op_ext(SERRC_EXIT,0xCCCC,"CORO driver value check_sum "
                    " verification failed",'I',NULL);
   }

   file_record(DE, 0, "CR", 0, NOHOLD);

   return;
}


/*-----------------------------------------------------------------*/
/* Function    : release_fa                                        */
/*                                                                 */
/* Description : release file address                              */
/*-----------------------------------------------------------------*/
void release_fa()
{
   struct stdhdr   rlcha_p;
   TESTBLK        *root;

   tx_begin();

   if ( (root = (TESTBLK *)find_record(DE, 0, "CR", 0, NOHOLD)) == 0)
   {
       serrc_op_ext(SERRC_EXIT,0xCCCC,"root record not found on DE "
                    " in release function",'I',NULL);
   }

   if ( root->fwd )
   {
      memcpy(rlcha_p.stdbid,(char *)&root->bwd,2);
      rlcha_p.stdrcc = 0;
      memcpy((char *)&rlcha_p.stdfad,
             (char *)&root->fwd,sizeof(rlcha_p.stdfad));
      rlcha(&rlcha_p);

      root->fwd = 0;
      root->fcnt = 0;
      root->value = 0;
      root->chksum = 0;
   }

   file_record(DE, 0, "CR", 0, NOHOLD);

   tx_commit();
}
